﻿------------------------------------------------------------------------------
-- GridStatusChainHealTarget by Napraptor
-- Credit goes to Slaren for GridStatusChainBest and Duugu for GridStatusCloseUp idea and original code
------------------------------------------------------------------------------

local QuickHealth = LibStub and LibStub("LibQuickHealth-2.0", true) -- don't error if not found
local MapFiles = LibStub("LibMapData-1.0") 
local GridStatus = Grid:GetModule("GridStatus")

GridStatusChainHealTarget = GridStatus:NewModule("GridStatusChainHealTarget")
GridStatusChainHealTarget.menuName = "Chain Heal Target"


-- constants

local ICON_CH = "Interface\\Icons\\Spell_Nature_HealingWaveGreater"


-- upvalues
local GridRoster = Grid:GetModule("GridRoster")
local GridFrame = Grid:GetModule("GridFrame")
local UnitGUID = UnitGUID
local GetPlayerMapPosition = GetPlayerMapPosition
local SetMapToCurrentZone = SetMapToCurrentZone
local GetCurrentMapDungeonLevel = GetCurrentMapDungeonLevel
local GetMapInfo = GetMapInfo
local UnitHealth = QuickHealth and QuickHealth.UnitHealth or UnitHealth
local UnitHealthMax = UnitHealthMax
local UnitIsVisible = UnitIsVisible
local UnitIsDeadOrGhost = UnitIsDeadOrGhost
local UnitIsConnected = UnitIsConnected 
local UnitIsEnemy = UnitIsEnemy
local UnitIsCharmed = UnitIsCharmed
local math_min = math.min
local math_floor = math.floor
local tinsert = table.insert
local tsort = table.sort

-- local data
local settings
local update_frame = CreateFrame("Frame")
local update_timer = 0
local ch_testrange_sq = 12.5 ^ 2

GridStatusChainHealTarget.defaultDB = {
	debug = false,

	alert_chainhealtarget = {
		enable = true,
		showjumps = true,
		color = { r = 1, g = 0, b = 0, a = 1 },
		color2 = { r = 1, g = 1, b = 0, a = 1 },
		priority = 99,
		range = true,
		cycle_time = 0.5,
		minjumps = 2,
		maxhealth = 85,
	},
}

local options = {
	["showjumps"] = {
		order = 100,
		type = "toggle",
		name = "Show jumps",
		desc = "If this is enabled, all targets that Chain Heal will jump to if casted on the suggested target, will be marked with the low priority color.",
		get = function () return GridStatusChainHealTarget.db.profile.alert_chainhealtarget.showjumps end,
		set = function (_, v) GridStatusChainHealTarget.db.profile.alert_chainhealtarget.showjumps = v end,
	},
	["color2"] = {
		order = 101,
		type = "color",
		name = "Jump color",
		desc = "The color used for Chain Heal jumps.",
		get = function () return GridStatusChainHealTarget.db.profile.alert_chainhealtarget.color2 end,
		set = function (_, v) GridStatusChainHealTarget.db.profile.alert_chainhealtarget.color2 = v end,
	},
	["frequency"] = {
		order = 102,
		type = "range",
		name = "Update frequency",
		desc = "Time between status refreshes, in seconds.",
		get = function () return GridStatusChainHealTarget.db.profile.alert_chainhealtarget.cycle_time end,
		set = function (_, v) GridStatusChainHealTarget.db.profile.alert_chainhealtarget.cycle_time = v end,
		min = 0.01,
		max = 1.5,
		step = 0.01,
	},
	["minjumps"] = {
		order = 103,
		type = "range",
		name = "Jump threshold",
		desc = "Only consider targets that will result in at least this number of Chain Heal jumps.",
		max = 3,
		min = 1,
		step = 1,
		get = function () return GridStatusChainHealTarget.db.profile.alert_chainhealtarget.minjumps end,
		set = function (_, v) GridStatusChainHealTarget.db.profile.alert_chainhealtarget.minjumps = v end,
	},
	["maxhealth"] = {
		order = 104,
		type = "range",
		name = "Health threshold",
		desc = "Only consider targets below this percentage of health.",
		max = 100,
		min = 1,
		step = 1,
		get = function () return GridStatusChainHealTarget.db.profile.alert_chainhealtarget.maxhealth end,
		set = function (_, v) GridStatusChainHealTarget.db.profile.alert_chainhealtarget.maxhealth = v end,
	},
}

function GridStatusChainHealTarget:OnInitialize()
	self.super.OnInitialize(self)
	
	self:RegisterStatus("alert_chainhealtarget", "Chain Heal Target", options, true)

	WorldMapFrame:Show()
	WorldMapFrame:Hide()
	
	settings = self.db.profile.alert_chainhealtarget
	self.debugging = self.db.profile.debug
end

function GridStatusChainHealTarget:OnStatusEnable(status)
	if status == "alert_chainhealtarget" then
		-- check talents the first frame, then change OnUpdate script to the real function
		update_frame:SetScript("OnUpdate",
			function()
				update_frame:SetScript("OnUpdate", function(_, elapsed) return self:OnUpdate(elapsed) end)
			end)
	end
end

function GridStatusChainHealTarget:OnStatusDisable(status)
	if status == "alert_chainhealtarget" then
		self.core:SendStatusLostAllUnits("alert_chainhealtarget")

		update_frame:SetScript("OnUpdate", nil)
	end
end

function GridStatusChainHealTarget:ClearChStatus()
  self.core:SendStatusLostAllUnits("alert_chainhealtarget")
end

function GridStatusChainHealTarget:OnUpdate(elapsed)
	update_timer = update_timer + elapsed

	if update_timer >= settings.cycle_time then
		update_timer = 0
		GridStatusChainHealTarget:RefreshAll()
	end
end

local refresh_state = {
	map_width = 0,
	map_height = 0,
	player_x = 0,
	player_y = 0,
	player_data = nil	
}

function GridStatusChainHealTarget:UnitsInRange(x1, y1, x2, y2, test_dist_sq)
	local xx = (x2 - x1) * refresh_state.map_width
	local yy = (y2 - y1) * refresh_state.map_height
	local dist_sq = xx*xx + yy*yy
	
	if dist_sq <= test_dist_sq then
		return true
	end
	return false
end

function GridStatusChainHealTarget:IsValidTarget(unitid)
	return not UnitIsDeadOrGhost(unitid) and
				UnitIsConnected(unitid) and 
				UnitIsVisible(unitid) and
				not (UnitIsCharmed(unitid) and UnitIsEnemy("player", unitid))
end

function GridStatusChainHealTarget:RefreshMapData()
	-- check player position
	refresh_state.player_x, refresh_state.player_y = GetPlayerMapPosition("player")
	if refresh_state.player_x <= 0 and refresh_state.player_y <= 0 then
		if WorldMapFrame:IsVisible() then
			return false
		end
		-- carbonite blinking map workaround
		local f = GetMouseFocus()
		while f do
			if f:GetName() == "NxMap1" then
				return false
			end
			f = f:GetParent()
		end

		SetMapToCurrentZone()
		refresh_state.player_x, refresh_state.player_y = GetPlayerMapPosition("player") 
	end
 	
	-- continue only if map supported
	if (refresh_state.player_x > 0 or refresh_state.player_y > 0) then
		local fileName = GetMapInfo()
		local currentLevel = GetCurrentMapDungeonLevel()
		refresh_state.map_width, refresh_state.map_height = MapFiles:MapArea(fileName, currentLevel) 
		if refresh_state.map_width ~= 0 and refresh_state.map_height ~= 0 then
			return true
		end
	end
	
	return false
end

function GridStatusChainHealTarget:RefreshAll()
	local refresh_state = refresh_state

	if not self:RefreshMapData() then
		self:ClearChStatus()
	else
		-- cache player data
		refresh_state.player_data = {}

		for cguid, unitid in GridRoster:IterateRoster() do
			local cx, cy = GetPlayerMapPosition(unitid)
			if (cx ~= 0 or cy ~= 0) and self:IsValidTarget(unitid) then
				local health = UnitHealth(unitid)
				local health_max = UnitHealthMax(unitid)
				
				local pdata = {
					guid = cguid,
					x = cx,
					y = cy,
					deficit = health_max - health,
					percent = health / health_max,
					inrange = {} -- list of unitids candidates to jump to
				}
	
				-- make list of players in range of everyone
				for tunitid, tdata in pairs(refresh_state.player_data) do
					-- Only include targets below maxhealth %
					if unitid ~= tunitid and self:UnitsInRange(pdata.x, pdata.y, tdata.x, tdata.y, ch_testrange_sq) then
						if tdata.percent < (settings.maxhealth * 0.01) then
							tinsert(pdata.inrange, tunitid)
						end
						if pdata.percent < (settings.maxhealth * 0.01) then
							tinsert(tdata.inrange, unitid)
						end
					end
				end
				
				refresh_state.player_data[unitid] = pdata
			end
		end


		self:RefreshCh()
	end
end

function GridStatusChainHealTarget:RefreshCh()
	local refresh_state = refresh_state
	
	local ch_best_uid = nil
	local ch_best_pdata = nil
	local ch_best_jumps = 0
	
	-- check all
	for unitid, p1 in pairs(refresh_state.player_data) do
		if not settings.range or GridFrame:UnitInRange(unitid, 40) then
			
			local curbest = nil
			local curbest_unitid = nil
			local curjumps = math_min(#(p1.inrange), 3)
			-- We're only interested in players that will generate at least minjumps number of jumps
			if curjumps >= settings.minjumps then

				for _, tunitid in pairs(p1.inrange) do
					local p2 = refresh_state.player_data[tunitid]
					if (not curbest) or (p2.deficit > curbest.deficit) then
						curbest = p2
						curbest_unitid = tunitid
					end
				end
		
				-- Check if we found a better target, based on number of jumps and health deficit
				if curbest and (curjumps >= ch_best_jumps) and ((not ch_best_pdata) or (curbest.deficit > ch_best_pdata.deficit)) then
					ch_best_pdata = curbest
					ch_best_uid = curbest_unitid
					ch_best_jumps = curjumps
				end
			end
		end
	end
	
	-- Clear and update status
	self:ClearChStatus()
	
	if ch_best_pdata then
		self.core:SendStatusGained(	ch_best_pdata.guid, 
						"alert_chainhealtarget",
						settings.priority,
						nil,
						settings.color,
						tostring(ch_best_jumps),
						1, 
						nil,
						ICON_CH)
		if settings.showjumps then
			for _, tunitid in pairs(ch_best_pdata.inrange) do
				if tunitid ~= ch_best_uid then
					local p1 = refresh_state.player_data[tunitid]
					self.core:SendStatusGained(	p1.guid, 
									"alert_chainhealtarget",
									settings.priority,
									nil,
									settings.color2,
									tostring(ch_best_jumps),
									1, 
									nil,
									ICON_CH)
				end
			end
		end
	end
end
